import { getPatterns } from './constants.js';
import { createContext } from './context.js';
import { defaultConfig } from './default-config.js';
import { parseChildren } from './parser/index.js';
import { headerFooter, headerFooterContent } from './render/header-footer.js';
import { simpleRender } from './render/simple-render.js';
import { htmlToDom } from './utils/html-to-dom.js';
export { simpleRender, getPatterns };
export const parse = (input, _config = defaultConfig()) => {
    const root = typeof input === 'string' ? htmlToDom(input) : input;
    if (root?.nodeName === 'TEMPLATE' || root?.nodeName === 'HTML' || root?.nodeName === '#document') {
        return parseTemplate(root, _config);
    }
    const ctx = createContext(_config);
    if (root) {
        const styles = root.querySelectorAll('style');
        const pageStyles = ctx.config.parseCss(styles);
        ctx.setPageStyles(pageStyles);
    }
    const content = root !== null ? parseChildren(root, ctx) : [];
    return {
        info: {
            title: 'PDF Document - created by html2pdfmake',
            creator: 'html2pdfmake',
            producer: 'html2pdfmake',
        },
        content: content,
        images: ctx.images,
        patterns: getPatterns(),
        pageSize: 'A4',
        pageOrientation: 'portrait',
        header: () => [],
        footer: () => [],
        pageMargins: [40, 60, 40, 60]
    };
};
export const parseTemplate = (template, _config = defaultConfig()) => {
    const ctx = createContext(_config);
    const root = template.nodeName === 'TEMPLATE'
        ? ctx.config.document().importNode(template.content, true)
        : template.nodeName === '#document' ? template.documentElement : template;
    const headers = headerFooterContent(root.querySelectorAll(':scope > header, body > header'), ctx);
    const footers = headerFooterContent(root.querySelectorAll(':scope > footer, body > footer'), ctx);
    const styles = root.querySelectorAll('style');
    const pageStyles = ctx.config.parseCss(styles);
    ctx.setPageStyles(pageStyles);
    const main = root.querySelector('main');
    const title = root.querySelector('title');
    const author = root.querySelector('meta[name="author"]');
    const subject = root.querySelector('meta[name="description"]');
    const keywords = root.querySelector('meta[name="keywords"]');
    const pageSize = root.querySelector('meta[name="page-size"]');
    const pageOrientation = root.querySelector('meta[name="page-orientation"]');
    const pageMargins = root.querySelector('meta[name="page-margins"]');
    const margins = pageMargins?.getAttribute('content')?.split(',').filter(m => m).map(m => parseFloat(m)) || [40, 60, 40, 60];
    const info = {
        title: title?.textContent || 'PDF Document - created by html2pdfmake',
        author: author?.getAttribute('content') || 'html2pdfmake',
        subject: subject?.getAttribute('content') || 'Created by html2pdfmake',
        keywords: keywords?.getAttribute('content') || 'html2pdfmake',
        creator: 'html2pdfmake',
        producer: 'html2pdfmake',
    };
    const content = main !== null ? parseChildren(main, ctx) : [];
    return {
        header(currentPage, pageCount) {
            return headerFooter(currentPage, pageCount, headers);
        },
        footer(currentPage, pageCount) {
            return headerFooter(currentPage, pageCount, footers);
        },
        info,
        content: content,
        images: ctx.images,
        patterns: getPatterns(),
        pageSize: pageSize?.getAttribute('content') || 'A4',
        pageOrientation: pageOrientation?.getAttribute('content') === 'landscape' ? 'landscape' : 'portrait',
        pageMargins: margins,
    };
};
