"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseElement = exports.parseByRule = exports.processItems = exports.getElementRule = exports.getNodeRule = exports.parseChildren = exports.isStackItem = exports.stackRegex = void 0;
const constants_js_1 = require("../constants.js");
const index_js_1 = require("../handler/index.js");
const index_js_2 = require("../props/index.js");
const collapse_js_1 = require("../utils/collapse.js");
const type_guards_js_1 = require("../utils/type-guards.js");
const unit_js_1 = require("../utils/unit.js");
const whitespace_js_1 = require("../utils/whitespace.js");
const parse_img_js_1 = require("./parse-img.js");
const parse_link_js_1 = require("./parse-link.js");
const parse_svg_js_1 = require("./parse-svg.js");
const parse_table_js_1 = require("./parse-table.js");
const parse_text_js_1 = require("./parse-text.js");
const parseAsHTMLCollection = (el) => ['TABLE', 'TBODY', 'TR', 'COLGROUP', 'COL', 'UL', 'OL', 'SELECT'].includes(el.nodeName) && 'children' in el;
exports.stackRegex = /^(address|blockquote|body|center|colgroup|dir|div|dl|fieldset|figure|figcaption|form|h[1-6]|hr|isindex|menu|noframes|noscript|ol|p|pre|table|ul|dd|dt|frameset|li|tbody|td|tfoot|th|thead|tr|html)$/i;
const isStackItem = (el) => exports.stackRegex.test(el.nodeName);
exports.isStackItem = isStackItem;
const parseChildren = (el, ctx, parentItem) => {
    const items = [];
    const children = parseAsHTMLCollection(el) ? el.children : el.childNodes;
    for (let i = 0; i < children.length; i++) {
        const item = (0, exports.parseByRule)(children[i], ctx, parentItem);
        if (item === null) {
            continue;
        }
        const isNewline = !!item[constants_js_1.META]?.[constants_js_1.IS_NEWLINE];
        const prevItem = items[items.length - 1];
        if (ctx.config.collapseMargin && prevItem) {
            (0, collapse_js_1.collapseMargin)(item, prevItem);
        }
        const prevIsBlock = prevItem && !('text' in prevItem);
        // Skip new lines
        if (isNewline && (items.length === 0 || !children[i + 1] || prevIsBlock)) {
            continue;
        }
        // Block item
        if (!('text' in item)) {
            items.push(item);
            continue;
        }
        const endWithNewLine = !!item[constants_js_1.META]?.[constants_js_1.END_WITH_NEWLINE];
        const startWithNewLine = !!item[constants_js_1.META]?.[constants_js_1.START_WITH_NEWLINE];
        const endWithWhiteSpace = !!item[constants_js_1.META]?.[constants_js_1.END_WITH_WHITESPACE];
        const startWithWhitespace = !!item[constants_js_1.META]?.[constants_js_1.START_WITH_WHITESPACE];
        const isWhitespace = !!item[constants_js_1.META]?.[constants_js_1.IS_WHITESPACE];
        const isAbsolutePosition = item?.[constants_js_1.META]?.[constants_js_1.STYLE]?.position === 'absolute';
        const textItem = Array.isArray(item.text)
            ? item
            : { text: [isNewline || isWhitespace ? (0, whitespace_js_1.addWhitespace)('newLine') : item] };
        if (!isNewline && !isWhitespace) {
            // https://developer.mozilla.org/en-US/docs/Web/API/Document_Object_Model/Whitespace
            // add whitespace before
            if (startWithNewLine || startWithWhitespace) {
                textItem.text.unshift((0, whitespace_js_1.addWhitespace)(startWithNewLine ? 'startWithNewLine' : 'startWithWhitespace'));
            }
            // add whitespace after
            if (endWithNewLine || endWithWhiteSpace) {
                textItem.text.push((0, whitespace_js_1.addWhitespace)(endWithNewLine ? 'endWithNewLine' : 'endWithWhiteSpace'));
            }
        }
        // Append text to last text element otherwise a new line is created
        if ((0, type_guards_js_1.isTextArray)(prevItem)) {
            if (ctx.config.collapseWhitespace) {
                (0, collapse_js_1.collapseWhitespace)(prevItem, textItem);
            }
            prevItem.text.push(textItem);
        }
        else if (isAbsolutePosition) {
            items.push(textItem);
        }
        else {
            // wrap so the next text items  will be appended to it
            items.push({
                text: [textItem]
            });
        }
    }
    return items;
};
exports.parseChildren = parseChildren;
const getNodeRule = (node) => {
    const nodeName = node.nodeName.toLowerCase();
    switch (nodeName) {
        case '#comment':
            return () => null;
        case '#text':
            return parse_text_js_1.parseText;
        default:
            return () => null;
    }
};
exports.getNodeRule = getNodeRule;
const getElementRule = (el) => {
    const nodeName = el.nodeName.toLowerCase();
    switch (nodeName) {
        case '#comment':
        case 'option': // see <select>
        case 'script':
        case 'style':
        case 'iframe':
        case 'object':
            return () => null;
        case '#text':
            return parse_text_js_1.parseText;
        case 'a':
            return (el) => {
                const href = el.getAttribute('href');
                if (!href) {
                    return (0, exports.parseElement)(el);
                }
                return (0, parse_link_js_1.parseLink)(href);
            };
        case 'br':
            return () => ({
                text: '\n'
            });
        case 'h1':
        case 'h2':
        case 'h3':
        case 'h4':
        case 'h5':
        case 'h6':
            return (el) => {
                const headline = (0, exports.parseElement)(el);
                if (headline === null) {
                    return null;
                }
                const meta = headline[constants_js_1.META] || {};
                meta[constants_js_1.POST_HANDLER] = index_js_1.handleHeadlineToc;
                headline[constants_js_1.META] = meta;
                return headline;
            };
        case 'qr-code': // CUSTOM
            return (el) => {
                const content = el.getAttribute('value');
                if (!content) {
                    return null;
                }
                const sizeAttr = el.getAttribute('data-size');
                const size = sizeAttr ? (0, unit_js_1.toUnit)(sizeAttr) : (0, unit_js_1.toUnit)('128px');
                return {
                    qr: content,
                    fit: size,
                };
            };
        case 'toc': // CUSTOM
            return (el) => {
                const content = el.textContent;
                if (!content) {
                    return null;
                }
                return {
                    toc: {
                        title: {
                            text: content,
                            bold: true,
                            fontSize: (0, unit_js_1.toUnit)('22px'),
                            margin: [0, 10, 0, 10]
                        },
                    },
                };
            };
        case 'table':
            return parse_table_js_1.parseTable;
        case 'ul':
            return () => {
                return {
                    ul: (items) => items
                };
            };
        case 'ol':
            return () => {
                return {
                    ol: (items) => items
                };
            };
        case 'img':
            return parse_img_js_1.parseImg;
        case 'svg':
            return parse_svg_js_1.parseSvg;
        case 'hr':
            // TODO find better <hr> alternative?
            return () => {
                return {
                    table: {
                        widths: ['*'],
                        body: [
                            [''],
                        ]
                    },
                    style: ['hr'],
                };
            };
        case 'input':
            // TODO acro-form
            return (el) => {
                if (el instanceof HTMLInputElement) {
                    return {
                        text: 'value' in el ? el.value : '',
                        [constants_js_1.META]: {
                            [constants_js_1.IS_INPUT]: true
                        }
                    };
                }
                return null;
            };
        case 'select':
            // TODO acro-form
            return (el) => {
                if (el instanceof HTMLSelectElement) {
                    const value = el.options[el.selectedIndex].value;
                    return {
                        text: value,
                        [constants_js_1.META]: {
                            [constants_js_1.IS_INPUT]: true
                        }
                    };
                }
                return null;
            };
        default:
            return exports.parseElement;
    }
};
exports.getElementRule = getElementRule;
const getItemByRule = (el, ctx) => {
    if (typeof ctx.config.nodeRule === 'function') {
        const result = ctx.config.nodeRule(el, ctx);
        if (result === null) {
            return null;
        }
        else if (result !== undefined) {
            return result;
        }
    }
    if ((0, type_guards_js_1.isElement)(el)) { // ELEMENT_NODE
        return (0, exports.getElementRule)(el)(el, ctx);
    }
    else if ((0, type_guards_js_1.isNode)(el)) { // TEXT_NODE || COMMENT_NODE
        return (0, exports.getNodeRule)(el)(el, ctx);
    }
    throw new Error('Unsupported Node Type: ' + el.nodeType);
};
const evaluateLazyChildren = (item, ctx) => {
    const el = item[constants_js_1.META][constants_js_1.NODE];
    if ('stack' in item && typeof item.stack === 'function') {
        const children = (0, exports.parseChildren)(el, ctx, item);
        item.stack = item.stack(children, ctx, item);
    }
    else if ('text' in item && typeof item.text === 'function') {
        const children = (0, exports.parseChildren)(el, ctx, item);
        item.text = item.text(children.filter(type_guards_js_1.isTextOrLeaf), ctx, item);
    }
    else if ('ul' in item && typeof item.ul === 'function') {
        const children = (0, exports.parseChildren)(el, ctx, item);
        item.ul = item.ul(children, ctx, item);
    }
    else if ('ol' in item && typeof item.ol === 'function') {
        const children = (0, exports.parseChildren)(el, ctx, item);
        item.ol = item.ol(children, ctx, item);
    }
    else if ('table' in item && typeof item.table.body === 'function') {
        const children = (0, exports.parseChildren)(el, ctx, item);
        item.table.body = item.table.body(children, ctx, item);
    }
};
const parseLazyChildren = (item, ctx, parentItem) => {
    if (typeof item === 'string') {
        return (0, index_js_1.preHandleLazyItem)(item);
    }
    const props = (0, index_js_2.computeProps)(item, ctx, parentItem);
    Object.assign(item, props);
    const preHandled = (0, index_js_1.preHandleLazyItem)(item);
    if (preHandled === null) {
        return null;
    }
    evaluateLazyChildren(preHandled, ctx);
    return preHandled;
};
const processItems = (item, ctx, parentItem) => {
    const withChildren = parseLazyChildren(item, ctx, parentItem);
    if (withChildren === null) {
        return null;
    }
    return (0, index_js_1.postHandleItem)(withChildren);
};
exports.processItems = processItems;
const parseByRule = (el, ctx, parentItem) => {
    const item = getItemByRule(el, ctx);
    if (item === null) {
        return null;
    }
    // Add ref to NODE
    const meta = item[constants_js_1.META] || {};
    meta[constants_js_1.NODE] = el;
    item[constants_js_1.META] = meta;
    return (0, exports.processItems)(item, ctx, parentItem);
};
exports.parseByRule = parseByRule;
const parseElement = (el) => {
    if ((0, exports.isStackItem)(el)) {
        return {
            stack: (items) => items
        };
    }
    return {
        text: (items) => {
            // Return flat
            if (items.length === 1 && 'text' in items[0] && Array.isArray(items[0].text)) {
                return items[0].text;
            }
            return items;
        }
    };
};
exports.parseElement = parseElement;
