"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.styleToProps = void 0;
const constants_js_1 = require("../constants.js");
const index_js_1 = require("../handler/index.js");
const color_js_1 = require("../utils/color.js");
const type_guards_js_1 = require("../utils/type-guards.js");
const unit_js_1 = require("../utils/unit.js");
const border_js_1 = require("./border.js");
const margin_js_1 = require("./margin.js");
const padding_js_1 = require("./padding.js");
const styleToProps = (item, ctx, styles, rootStyles = {}) => {
    const props = {
        [constants_js_1.META]: {
            ...(item[constants_js_1.META] || {}),
            [constants_js_1.STYLE]: {},
            ...(item[constants_js_1.META]?.[constants_js_1.STYLE] || {}),
        }
    };
    const meta = props[constants_js_1.META];
    const image = (0, type_guards_js_1.isImage)(item);
    const table = (0, type_guards_js_1.isTable)(item);
    const text = (0, type_guards_js_1.isTextSimple)(item);
    const list = (0, type_guards_js_1.isList)(item);
    const rootFontSize = (0, unit_js_1.toUnit)(rootStyles['font-size'] || '16px');
    Object.keys(styles).forEach((key) => {
        const directive = key;
        const value = ('' + styles[key]).trim();
        props[constants_js_1.META][constants_js_1.STYLE][directive] = value;
        if (typeof ctx.config.styleRule === 'function' && ctx.config.styleRule(directive, value, props)) {
            return;
        }
        switch (directive) {
            case 'padding': {
                const paddings = (0, unit_js_1.expandValueToUnits)(value);
                if (table && paddings !== null) {
                    let layout = props.layout || item.layout || {};
                    if (typeof layout === 'string') {
                        layout = {};
                    }
                    layout.paddingLeft = () => Number(paddings[constants_js_1.POS_LEFT]);
                    layout.paddingRight = () => Number(paddings[constants_js_1.POS_RIGHT]);
                    layout.paddingTop = (i) => (i === 0) ? Number(paddings[constants_js_1.POS_TOP]) : 0;
                    layout.paddingBottom = (i, node) => (i === node.table.body.length - 1) ? Number(paddings[constants_js_1.POS_BOTTOM]) : 0;
                    props.layout = layout;
                }
                else if (paddings !== null) {
                    (0, padding_js_1.computePadding)(item, props, 'padding-top', Number(paddings[constants_js_1.POS_TOP]));
                    (0, padding_js_1.computePadding)(item, props, 'padding-left', Number(paddings[constants_js_1.POS_LEFT]));
                    (0, padding_js_1.computePadding)(item, props, 'padding-right', Number(paddings[constants_js_1.POS_RIGHT]));
                    (0, padding_js_1.computePadding)(item, props, 'padding-bottom', Number(paddings[constants_js_1.POS_BOTTOM]));
                }
                break;
            }
            case 'padding-left':
            case 'padding-top':
            case 'padding-right':
            case 'padding-bottom':
                (0, padding_js_1.computePadding)(item, props, directive, value);
                break;
            case 'border':
            case 'border-bottom':
            case 'border-top':
            case 'border-right':
            case 'border-left':
                (0, border_js_1.computeBorder)(item, props, directive, value);
                break;
            case 'font-size': {
                props.fontSize = (0, unit_js_1.toUnit)(value, rootFontSize);
                break;
            }
            case 'line-height':
                if (value !== 'inherit') { // TODO handle inherit
                    props.lineHeight = (0, unit_js_1.toUnit)(value, rootFontSize);
                }
                break;
            case 'letter-spacing':
                props.characterSpacing = (0, unit_js_1.toUnit)(value);
                break;
            case 'text-align':
                if (['left', 'right', 'justify', 'center'].includes(value)) {
                    props.alignment = value;
                }
                break;
            case 'font-feature-settings': {
                const settings = value.split(',').filter(s => s).map(s => s.replace(/['"]/g, ''));
                const fontFeatures = item.fontFeatures || props.fontFeatures || [];
                fontFeatures.push(...settings);
                props.fontFeatures = fontFeatures;
                break;
            }
            case 'font-weight':
                switch (value) {
                    case 'bold':
                        props.bold = true;
                        break;
                    case 'normal':
                        props.bold = false;
                        break;
                }
                break;
            case 'text-decoration':
                switch (value) {
                    case 'underline':
                        props.decoration = 'underline';
                        break;
                    case 'line-through':
                        props.decoration = 'lineThrough';
                        break;
                    case 'overline':
                        props.decoration = 'overline';
                        break;
                }
                break;
            case 'text-decoration-color':
                props.decorationColor = (0, color_js_1.parseColor)(value);
                break;
            case 'text-decoration-style':
                if (['dashed', 'dotted', 'double', 'wavy'].includes(value)) {
                    props.decorationStyle = value;
                }
                break;
            case 'vertical-align':
                if (value === 'sub') {
                    props.sub = true;
                }
                break;
            case 'font-style':
                switch (value) {
                    case 'italic':
                        props.italics = true;
                        break;
                }
                break;
            case 'font-family': {
                const font = value.split(',').filter(f => !!f).map(f => f.replace(/["']/g, '').trim());
                if (ctx.fonts) {
                    props.font = Object.keys(ctx.fonts).find(f => font.includes(f)) || ctx.config.defaultFont;
                }
                else {
                    props.font = font[0] || ctx.config.defaultFont;
                }
                break;
            }
            case 'color':
                props.color = (0, color_js_1.parseColor)(value);
                break;
            case 'background':
            case 'background-color':
                if (table) {
                    let layout = item.layout || {};
                    if (typeof layout === 'string') {
                        layout = {};
                    }
                    layout.fillColor = () => (0, color_js_1.parseColor)(value);
                    props.layout = layout;
                }
                else if ((0, type_guards_js_1.isTdOrTh)(item)) {
                    props.fillColor = (0, color_js_1.parseColor)(value);
                }
                else {
                    props.background = ['fill', (0, color_js_1.parseColor)(value)];
                }
                break;
            case 'margin': {
                const margin = (0, unit_js_1.expandValueToUnits)(value)?.map(value => typeof value === 'string' ? 0 : value);
                if (margin) {
                    (0, margin_js_1.computeMargin)(props, item, margin[constants_js_1.POS_TOP], constants_js_1.POS_TOP);
                    (0, margin_js_1.computeMargin)(props, item, margin[constants_js_1.POS_LEFT], constants_js_1.POS_LEFT);
                    (0, margin_js_1.computeMargin)(props, item, margin[constants_js_1.POS_RIGHT], constants_js_1.POS_RIGHT);
                    (0, margin_js_1.computeMargin)(props, item, margin[constants_js_1.POS_BOTTOM], constants_js_1.POS_BOTTOM);
                }
                break;
            }
            case 'margin-left':
                (0, margin_js_1.computeMargin)(props, item, (0, unit_js_1.toUnit)(value), constants_js_1.POS_LEFT);
                break;
            case 'margin-top':
                (0, margin_js_1.computeMargin)(props, item, (0, unit_js_1.toUnit)(value), constants_js_1.POS_TOP);
                break;
            case 'margin-right':
                (0, margin_js_1.computeMargin)(props, item, (0, unit_js_1.toUnit)(value), constants_js_1.POS_RIGHT);
                break;
            case 'margin-bottom':
                (0, margin_js_1.computeMargin)(props, item, (0, unit_js_1.toUnit)(value), constants_js_1.POS_BOTTOM);
                break;
            case 'page-break-before':
                if (value === 'always') {
                    props.pageBreak = 'before';
                }
                break;
            case 'page-break-after':
                if (value === 'always') {
                    props.pageBreak = 'after';
                }
                break;
            case 'position': {
                // TODO supoprt bottom and right
                const x = (0, unit_js_1.toUnit)(styles['left'] || 0);
                const y = (0, unit_js_1.toUnit)(styles['top'] || 0);
                if (value === 'absolute') {
                    props.absolutePosition = {
                        x,
                        y
                    };
                }
                else if (value === 'relative') {
                    props.relativePosition = {
                        x,
                        y
                    };
                }
                break;
            }
            case 'white-space':
                if (value === 'pre' && meta[constants_js_1.NODE]) {
                    if (text) {
                        props.text = meta[constants_js_1.NODE]?.textContent || '';
                    }
                    props.preserveLeadingSpaces = true;
                }
                break;
            case 'display':
                if (value === 'flex') {
                    props[constants_js_1.META][constants_js_1.POST_HANDLER] = index_js_1.handleColumns;
                }
                else if (value === 'none') {
                    props[constants_js_1.META][constants_js_1.PRE_HANDLER] = () => null;
                }
                break;
            case 'opacity':
                props.opacity = Number(parseFloat(value));
                break;
            case 'gap':
                props.columnGap = (0, unit_js_1.toUnit)(value) / 2; // TODO check
                break;
            case 'list-style-type':
            case 'list-style':
                if (list) {
                    props.type = value;
                }
                else {
                    props.listType = value;
                }
                break;
            case 'width':
                if (table) {
                    const width = (0, unit_js_1.toUnitOrValue)(value);
                    if (width !== null) {
                        item.table.widths = [width];
                    }
                }
                else if (image) {
                    props.width = (0, unit_js_1.toUnit)(value);
                }
                break;
            case 'height':
                if (image) {
                    props.height = (0, unit_js_1.toUnit)(value);
                }
                break;
            case 'max-height':
                if (image) {
                    props.maxHeight = (0, unit_js_1.toUnit)(value);
                }
                break;
            case 'max-width':
                if (image) {
                    props.maxWidth = (0, unit_js_1.toUnit)(value);
                }
                break;
            case 'min-height':
                if (image) {
                    props.minHeight = (0, unit_js_1.toUnit)(value);
                }
                break;
            case 'min-width':
                if (image) {
                    props.minWidth = (0, unit_js_1.toUnit)(value);
                }
                break;
            case 'object-fit':
                if (value === 'contain' && image) {
                    meta[constants_js_1.POST_HANDLER] = index_js_1.handleImg;
                }
                break;
            default:
                break;
        }
    });
    return props;
};
exports.styleToProps = styleToProps;
